/************************************************************************
 * libidmef: A library to create IDMEF messages in XML format.
 * Author: Joe McAlerney, Silicon Defense, (joey@SiliconDefense.com)
 *
 * Copyright (c) 2000 by Silicon Defense (http://www.silicondefense.com/)
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 **************************************************************************/

#include "idmefxml.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

/*************************************************************************
 * With the exception of newSimpleElement, each of these "new" functions 
 * accepts an xmlNodePtr, and a variable list of parameters, which should be
 * xmlNodePtrs as well.  The parameters must be passed in the order specified
 * by the DTD.  The functions will check to see if all required parameters 
 * were passed.  If not, it will return NULL.  Possible parameters are
 * displayed above the functions.
 *
 * Notes: At least ONE parameter is required in each function.  This parameter
 *        is labled 'first'. 
 *
 *        You must pass NULL in as the last argument. This is to let the 
 *        function know when to quit reading in arguments.  Failure to do this
 *        will result in "undefined behavior".
 *************************************************************************/
 
/*==================================================================
 * Name: newIDMEF_Message
 *
 * Possible parameters 
 *
 * ATTRIBUTES
 * xmlNodePtr version: optional
 *
 * SUB-ELEMENTS
 * xmlNodePtr Alert    : zero or more
 * xmlNodePtr Heartbeat: zero or more
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newIDMEF_Message(xmlNodePtr first, ...)
{
  xmlNodePtr p, theRoot;
  va_list ap;
  xmlChar *tmp_content = NULL;


  if(doc == NULL)        /* have we called initialize successfully yet? */
  {
      fprintf(stderr,"idmefxml.c: newIDMEF_Message(): you must call createCurrentDoc first\n");
      return NULL;
  }

  theRoot = xmlDocGetRootElement(doc);
  
  /* Set the IDMEF Version attribute */
  xmlSetProp(theRoot, "version", IDMEF_MESSAGE_VERSION);

  if(first == NULL) // do we have at least one parameter?
    return theRoot;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);
       
     switch(*(p->name))
     {
       /*
          case 'v':
          if(strcmp(p->name, "version") == 0)
	  {   
 	     xmlSetProp(theRoot, "version", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"IDMEF-Message");
          break;
       */
        case 'A':
          if(strcmp(p->name, "Alert") == 0)
  	    xmlAddChild(theRoot, p);
          else
            badNode(p,"IDMEF-Message");
          break;
        case 'H':
          if(strcmp(p->name, "Heartbeat") == 0)
 	     xmlAddChild(theRoot, p);
          else
            badNode(p,"IDMEF-Message");
          break;
        default:
          badNode(p,"IDMEF-Message");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(theRoot);
}

/*==================================================================
 * Name: newAlert
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr alertid: required
 * xmlNodePtr impact : required
 * xmlNodePtr version: required
 *
 * SUB-ELEMENTS
 * xmlNodePtr Time            : exactly one
 * xmlNodePtr Analyzer        : exactly one
 * xmlNodePtr Classification  : one or more
 * xmlNodePtr Source          : zero or more
 * xmlNodePtr Target          : zero or more
 * xmlNodePtr ToolAlert       : zero or one
 * xmlNodePtr OverflowAlert   : zero or one
 * xmlNodePtr CorrelationAlert: zero or one
 * xmlNodePtr AdditionalData  : zero or more
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newAlert(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Alert");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAlert(): xmlNewNode(NULL,\"Alert\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;
       
     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'a':
          if(strcmp(p->name, "alertid") == 0)
	  {
      	     xmlSetProp(cur, "alertid", tmp_content);
             xmlFreeNode(p);
          }         
          else
            badNode(p,"Alert");
          break;
        case 'i':
          if(strcmp(p->name, "impact") == 0)
	  {
 	     xmlSetProp(cur, "impact", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Alert");
          break;
        case 'v':
          if(strcmp(p->name, "version") == 0)
	  {
 	     xmlSetProp(cur, "version", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Alert");
          break;
        case 'T':
          switch(*(p->name + 1))
	  {
     	     case 'i':
               if(strcmp(p->name, "Time") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
     	     case 'a':
               if(strcmp(p->name, "Target") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
     	     case 'o':
               if(strcmp(p->name, "ToolAlert") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
  	     default:
               badNode(p,"Alert");
	  }
          break;
        case 'A':
          switch(*(p->name + 1))
	  {
     	     case 'n':
               if(strcmp(p->name, "Analyzer") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
     	     case 'd':
               if(strcmp(p->name, "AdditionalData") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
    	     default:
               badNode(p,"Alert");   
          }
          break;
        case 'C':
          switch(*(p->name + 1))
	  {
     	     case 'l':
               if(strcmp(p->name, "Classification") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
     	     case 'o':
               if(strcmp(p->name, "CorrelationAlert") == 0)
  	          xmlAddChild(cur, p);
               else
                  badNode(p,"Alert");
               break;
    	     default:
               badNode(p,"Alert");   
          }
          break;
        case 'S':
          if(strcmp(p->name, "Source") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Alert");
          break;
        case 'O':
          if(strcmp(p->name, "OverflowAlert") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Alert");
          break;
        default:
          badNode(p,"Alert");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newHeartbeat
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr heartbeatid: required
 *
 * SUB-ELEMENTS
 * xmlNodePtr Time            : exactly one
 * xmlNodePtr Analyzer        : exactly one
 * xmlNodePtr AdditionalData  : zero or more
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newHeartbeat(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Heartbeat");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newHeartbeat(): xmlNewNode(NULL,\"Heartbeat\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
  	case 'h': /* heartbeatid */
          if(strcmp(p->name, "heartbeatid") == 0)
	  {
	    xmlSetProp(cur, "heartbeatid", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Heartbeat");
          break;
        case 'T': /* Time */
          if(strcmp(p->name, "Time") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Heartbeat");
          break;
        case 'A':
          switch(*(p->name + 1))
	  {
  	     case 'n': /* Analyzer */
	       if(strcmp(p->name, "Analyzer") == 0)
  	         xmlAddChild(cur, p);
	       else
		 badNode(p,"Heartbeat");
	       break;
	     case 'd': /* AdditionalData */
	       if(strcmp(p->name, "AdditionalData") == 0)
  	         xmlAddChild(cur, p);
	       else
		 badNode(p,"Heartbeat");
	       break;
             default:
               badNode(p,"Heartbeat");
	  }
          break;
        default:
          badNode(p,"Heartbeat");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newToolAlert
 *
 * Possible parameters 
 * 
 * SUB-ELEMENTS
 * xmlNodePtr name   : exactly one
 * xmlNodePtr command: zero or more
 * xmlNodePtr alertid: zero or more
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newToolAlert(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;

  cur = xmlNewNode(NULL,"ToolAlert");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newToolAlert: xmlNewNode(NULL,\"ToolAlert\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     switch(*(p->name))
     {
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"ToolAlert");
          break;
        case 'c': /* command */
          if(strcmp(p->name, "command") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"ToolAlert");
          break;
        case 'a': /* alertid */
          if(strcmp(p->name, "alertid") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"ToolAlert");
          break;
        default:
          badNode(p,"ToolAlert");
     }
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newOverflowAlert
 *
 * Possible parameters 
 * 
 * SUB-ELEMENTS
 * xmlNodePtr program: exactly one
 * xmlNodePtr size   : exactly one
 * xmlNodePtr buffer : zero or one
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newOverflowAlert(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;

  cur = xmlNewNode(NULL,"OverflowAlert");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newOverflowAlert: xmlNewNode(NULL,\"OverflowAlert\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     switch(*(p->name))
     {
        case 'p': /* program */
          if(strcmp(p->name, "program") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"OverflowAlert");
          break;
        case 's': /* size */
          if(strcmp(p->name, "size") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"OverflowAlert");
          break;
        case 'b': /* buffer */
          if(strcmp(p->name, "buffer") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"OverflowAlert");
          break;
        default:
          badNode(p,"OverflowAlert");
     }
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newCorrelationAlert
 *
 * Possible parameters 
 * 
 * SUB-ELEMENTS
 * xmlNodePtr alertid: one or more
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newCorrelationAlert(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;

  cur = xmlNewNode(NULL,"CorrelationAlert");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newCorrelationAlert: xmlNewNode(NULL,\"CorrelationAlert\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     switch(*(p->name))
     {
        case 'a': /* alertid */
          if(strcmp(p->name, "alertid") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"CorrelationAlert");
          break;
        default:
          badNode(p,"CorrelationAlert");
     }
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newTime
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr offset: optional
 *
 * SUB-ELEMENTS
 * xmlNodePtr ntpstamp    : exactly one
 * xmlNodePtr date        : exactly one
 * xmlNodePtr time        : exactly one
 * xmlNodePtr DetectTime  : zero or one
 * xmlNodePtr AnalyzerTime: zero or one
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newTime(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Time");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newTime: xmlNewNode(NULL,\"Time\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
  	case 'o': /* offset */
          if(strcmp(p->name, "offset") == 0)
	  {
	     xmlSetProp(cur, "offset", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Time");
          break;
        case 'n': /* ntpstamp */
          if(strcmp(p->name, "ntpstamp") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Time");
          break;
        case 'd': /* date */
          if(strcmp(p->name, "date") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Time");
          break;
        case 't': /* time */
          if(strcmp(p->name, "time") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Time");
          break;
        case 'D': /* DetectTime */
          if(strcmp(p->name, "DetectTime") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Time");
          break;
        case 'A': /* AnalyzerTime */
          if(strcmp(p->name, "AnalyzerTime") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Time");
          break;
        default:
          badNode(p,"Time");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newAnalyzerTime
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr offset: optional
 *
 * SUB-ELEMENTS
 * xmlNodePtr ntpstamp    : exactly one
 * xmlNodePtr date        : exactly one
 * xmlNodePtr time        : exactly one
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newAnalyzerTime(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"AnalyzerTime");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAnalyzerTime: xmlNewNode(NULL,\"AnalyzerTime\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
  	case 'o': /* offset */
          if(strcmp(p->name, "offset") == 0)
	  {
	     xmlSetProp(cur, "offset", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"AnalyzerTime");
          break;
        case 'n': /* ntpstamp */
          if(strcmp(p->name, "ntpstamp") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"AnalyzerTime");
          break;
        case 'd': /* date */
          if(strcmp(p->name, "date") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"AnalyzerTime");
          break;
        case 't': /* time */
          if(strcmp(p->name, "time") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"AnalyzerTime");
          break;
        default:
          badNode(p,"AnalyzerTime");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newDetectTime
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr offset: optional
 *
 * SUB-ELEMENTS
 * xmlNodePtr ntpstamp    : exactly one
 * xmlNodePtr date        : exactly one
 * xmlNodePtr time        : exactly one
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newDetectTime(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"DetectTime");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newDetectTime: xmlNewNode(NULL,\"DetectTime\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
  	case 'o': /* offset */
          if(strcmp(p->name, "offset") == 0)
	  {
	     xmlSetProp(cur, "offset", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"DetectTime");
          break;
        case 'n': /* ntpstamp */
          if(strcmp(p->name, "ntpstamp") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"DetectTime");
          break;
        case 'd': /* date */
          if(strcmp(p->name, "date") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"DetectTime");
          break;
        case 't': /* time */
          if(strcmp(p->name, "time") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"DetectTime");
          break;
        default:
          badNode(p,"DetectTime");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newTarget
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr targetid: optional (zero or one?)
 * xmlNodePtr decoy   : required (exactly one?)
 *
 * SUB-ELEMENTS
 * xmlNodePtr Node   : zero or one
 * xmlNodePtr User   : zero or one
 * xmlNodePtr Process: zero or one
 * xmlNodePtr Service: zero or one
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newTarget(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Target");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newTarget: xmlNewNode(NULL,\"Target\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
  	case 't': /* targetid */
          if(strcmp(p->name, "targetid") == 0)
	  {
	     xmlSetProp(cur, "targetid", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Target");
          break;
	case 'd': /* decoy */
	  if(strcmp(p->name, "decoy") == 0)
	  {
	     xmlSetProp(cur, "decoy", tmp_content);
             xmlFreeNode(p);
          }
	  else
	    badNode(p,"Target");
	  break;
        case 'N': /* Node */
	  if(strcmp(p->name, "Node") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Target");
          break;
        case 'U': /* User */
	  if(strcmp(p->name, "User") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Target");
          break;
        case 'P': /* Process */
          if(strcmp(p->name, "Process") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Target");
          break;
        case 'S': /* Service */
          if(strcmp(p->name, "Service") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Target");
          break;
        default:
          badNode(p,"Target");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newSource
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr sourceid: optional (zero or one?)
 * xmlNodePtr spoofed : required (exactly one?)
 *
 * SUB-ELEMENTS
 * xmlNodePtr Node   : zero or one
 * xmlNodePtr User   : zero or one
 * xmlNodePtr Process: zero or one
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newSource(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Source");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newSource: xmlNewNode(NULL,\"Source\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 's': /* ident */
          switch(*(p->name + 1))
	  {
  	     case 'o': /* sourceid */
	       if(strcmp(p->name, "sourceid") == 0)
	       {
		 xmlSetProp(cur, "sourceid", tmp_content);
		 xmlFreeNode(p);
               }
	       else
		 badNode(p,"Source");
	       break;
	     case 'p': /* spoofed */
	       if(strcmp(p->name, "spoofed") == 0)
	       {
		 xmlSetProp(cur, "spoofed", tmp_content);
                 xmlFreeNode(p);
	       }
	       else
		 badNode(p,"Source");
	       break;
             default:
               badNode(p,"Source");
	  }
          break;
        case 'N': /* Node */
	  if(strcmp(p->name, "Node") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Source");
          break;
        case 'U': /* User */
	  if(strcmp(p->name, "User") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Source");
          break;
        case 'P': /* Process */
          if(strcmp(p->name, "Process") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Source");
          break;
        default:
          badNode(p,"Source");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newAnalyzer
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr ident: required (exactly one?)
 *
 * SUB-ELEMENTS
 * xmlNodePtr Node   : zero or one
 * xmlNodePtr Process: zero or one
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newAnalyzer(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Analyzer");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAnalyzer: xmlNewNode(NULL,\"Analyzer\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Analyzer");
          break;
        case 'N': /* Node */
	  if(strcmp(p->name, "Node") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Analyzer");
          break;
        case 'P': /* Process */
          if(strcmp(p->name, "Process") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Analyzer");
          break;
        default:
          badNode(p,"Analyzer");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newUser
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr ident   : optional (zero or one?)
 * xmlNodePtr category: required (exactly one?)
 *
 * SUB-ELEMENTS
 * xmlNodePtr name   : zero or one if <uid> info is provided;
 *                     exactly one otherwise
 * xmlNodePtr uid    : zero or one if <name> info is provided;
 *                     exactly one otherwise
 * xmlNodePtr group  : zero or one
 * xmlNodePtr gid    : zero or one
 * xmlNodePtr serial : zero or one
 * xmlNodePtr Address: zero or more
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newUser(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"User");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newUser: xmlNewNode(NULL,\"User\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"User");
          break;
        case 'c': /* category */
          if(strcmp(p->name, "category") == 0)
	  {
 	     xmlSetProp(cur, "category", tmp_content);
             xmlFreeNode(p);
          }          
          else
            badNode(p,"User");
          break;
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"User");
          break;
        case 'u': /* uid */
	  if(strcmp(p->name, "uid") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"User");
          break;
        case 'g':
          switch(*(p->name + 1))
	  {
  	     case 'r': /* group */
	       if(strcmp(p->name, "group") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"User");
	       break;
	     case 'i': /* gid */
	       if(strcmp(p->name, "gid") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"User");
	       break;
             default:
               badNode(p,"User");
	  }
          break;
        case 's': /* serial */
	  if(strcmp(p->name, "serial") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"User");
          break;
        case 'A': /* Address */
          if(strcmp(p->name, "Address") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"User");
          break;
        default:
          badNode(p,"User");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newService
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr ident: required (exactly one?)
 *
 * SUB-ELEMENTS
 * xmlNodePtr name       : zero or one
 * xmlNodePtr dport      : zero or one
 * xmlNodePtr sport      : zero or one
 * xmlNodePtr protocol   : zero or one
 * xmlNodePtr portlist   : zero or one
 * xmlNodePtr SNMPService: zero or one
 * xmlNodePtr WebService : zero or one
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newService(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Service");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newService: xmlNewNode(NULL,\"Service\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Service");
          break;
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Service");
          break;
        case 'd': /* dport */
	  if(strcmp(p->name, "dport") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Service");
          break;
        case 's': /* sport */
	  if(strcmp(p->name, "sport") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"Service");
          break;
        case 'p':
          switch(*(p->name + 1))
	  {
  	     case 'r': /* protocol */
	       if(strcmp(p->name, "protocol") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"Service");
	       break;
	     case 'o': /* portlist */
	       if(strcmp(p->name, "portlist") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"Service");
	       break;
             default:
               badNode(p,"Service");
	  }
          break;
        case 'S': /* SNMPService */
          if(strcmp(p->name, "SNMPService") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Service");
          break;
        case 'W': /* WebService */
          if(strcmp(p->name, "WebService") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"Service");
          break;
        default:
          badNode(p,"Service");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newWebService
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr ident: optional (zero or one?)
 *
 * SUB-ELEMENTS
 * xmlNodePtr url      : exactly one
 * xmlNodePtr cgi      : zero or one
 * xmlNodePtr method   : zero or one
 * xmlNodePtr Arguments: zero or one
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newWebService(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"WebService");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newWebService: xmlNewNode(NULL,\"WebService\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {          
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"WebService");
          break;
        case 'u': /* url */
          if(strcmp(p->name, "url") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"WebService");
          break;
        case 'c': /* cgi */
	  if(strcmp(p->name, "cgi") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"WebService");
          break;
        case 'm': /* method */
	  if(strcmp(p->name, "method") == 0)
	    xmlAddChild(cur, p);
          else
	    badNode(p,"WebService");
          break;
        case 'A': /* Arguments */
          if(strcmp(p->name, "Arguments") == 0)
       	    xmlAddChild(cur, p);
          else
            badNode(p,"WebService");
          break;
        default:
          badNode(p,"WebService");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newSNMPService
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr ident: optional (zero or one?)
 *
 * SUB-ELEMENTS
 * xmlNodePtr oid      : zero or one
 * xmlNodePtr community: zero or one
 * xmlNodePtr command  : zero or one
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newSNMPService(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"SNMPService");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newSNMPService: xmlNewNode(NULL,\"SNMPService\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"SNMPService");
          break;
        case 'o': /* oid */
          if(strcmp(p->name, "oid") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"SNMPService");
          break;
        case 'c':
          switch(*(p->name + 4))
	  {
  	     case 'u': /* community */
	       if(strcmp(p->name, "community") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"SNMPService");
	       break;
	     case 'a': /* command */
	       if(strcmp(p->name, "command") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"SNMPService");
	       break;
             default:
               badNode(p,"SNMPService");
	  }
          break;
        default:
          badNode(p,"SNMPService");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newProcess
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr ident: optional (zero or one?)
 *
 * SUB-ELEMENTS
 * xmlNodePtr name       : exactly one
 * xmlNodePtr pid        : zero or one
 * xmlNodePtr path       : zero or one
 * xmlNodePtr Arguments  : zero or one
 * xmlNodePtr Environment: zero or more
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newProcess(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Process");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newProcess: xmlNewNode(NULL,\"Process\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Process");
          break;
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Process");
          break;
        case 'p':
          switch(*(p->name + 1))
	  {
  	     case 'i': /* pid */
	       if(strcmp(p->name, "pid") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"Process");
	       break;
	     case 'a': /* path */
	       if(strcmp(p->name, "path") == 0)
		 xmlAddChild(cur, p);
	       else
		 badNode(p,"Process");
	       break;
             default:
               badNode(p,"Process");
	  }
          break;
        case 'A': /* Arguments */
          if(strcmp(p->name, "Arguments") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Process");
          break;
        case 'E': /* Environment */
          if(strcmp(p->name, "Environment") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Process");
          break;
        default:
          badNode(p,"Process");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newNode
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr ident: optional (zero or one?)
 * xmlNodePtr category: optional (zero or one?)
 *
 * SUB-ELEMENTS
 * xmlNodePtr name    : zero or one
 * xmlNodePtr location: zero or one
 * xmlNodePtr Address : zero or more
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newNode(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Node");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newNode: xmlNewNode(NULL,\"Node\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Node");
          break;
        case 'c': /* category */
          if(strcmp(p->name, "category") == 0)
	  {
 	     xmlSetProp(cur, "category", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Node");
          break;
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Node");
          break;
        case 'l': /* location */
          if(strcmp(p->name, "location") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Node");
          break;
        case 'A': /* Address */
          if(strcmp(p->name, "Address") == 0)
            xmlAddChild(cur,p);
          else
            badNode(p,"Node");
          break;
        default:
          badNode(p,"Node");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newClassification
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr origin: optional (zero or one?)
 *
 * SUB-ELEMENTS
 * xmlNodePtr name: required (at least one?)
 * xmlNodePtr url : required (at least one?)
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newClassification(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Classification");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newClassification: xmlNewNode(NULL,\"Classification\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'o': /* origin */
          if(strcmp(p->name, "origin") == 0)
	  {
 	     xmlSetProp(cur, "origin", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Classification");
          break;
        case 'n': /* name */
          if(strcmp(p->name, "name") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Classification");
          break;
        case 'u': /* url */
          if(strcmp(p->name, "url") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Classification");
          break;
        default:
          badNode(p,"Classification");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);
}

/*==================================================================
 * Name: newAddress
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr ident   : optional (zero or one?)
 * xmlNodePtr category: required (at least one?)
 *
 * SUB-ELEMENTS
 * xmlNodePtr address: required (at least one?)
 * xmlNodePtr netmask: optional (zero or one?)
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newAddress(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"Address");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAddress: xmlNewNode(NULL,\"Address\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'i': /* ident */
          if(strcmp(p->name, "ident") == 0)
	  {
 	     xmlSetProp(cur, "ident", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Address");
          break;
        case 'c': /* category */
          if(strcmp(p->name, "category") == 0)
	  {
 	     xmlSetProp(cur, "category", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"Address");
          break;
        case 'a': /* address */
          if(strcmp(p->name, "address") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Address");
          break;
        case 'n': /* netmask */
          if(strcmp(p->name, "netmask") == 0)
  	    xmlAddChild(cur, p);
          else
            badNode(p,"Address");
          break;
        default:
          badNode(p,"Address");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);

}

/*==================================================================
 * Name: newAdditionalData
 *
 * Possible parameters 
 * 
 * ATTRIBUTES
 * xmlNodePtr meaning: optional (zero or one?)
 * xmlNodePtr type   : optional (zero or one?)
 *
 * VALUE
 * xmlNodePtr value: exactly one (the additional data)
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newAdditionalData(xmlNodePtr first, ...)
{
  xmlNodePtr cur, p;
  va_list ap;
  xmlChar *tmp_content = NULL;

  cur = xmlNewNode(NULL,"AdditionalData");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAdditionalData: xmlNewNode(NULL,\"AdditionalData\") failed\n");
    return NULL;
  }

  if(first == NULL) // do we have at least one parameter?
    return cur;

  // cycle through the arguments
  va_start(ap, first);

  for(p = first; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name == NULL)
       return NULL;

     tmp_content = xmlNodeGetContent(p);

     switch(*(p->name))
     {
        case 'm': /* meaning */
          if(strcmp(p->name, "meaning") == 0)
	  {
 	     xmlSetProp(cur, "meaning", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"AdditionalData");
          break;
        case 't': /* type */
          if(strcmp(p->name, "type") == 0)
	  {
 	     xmlSetProp(cur, "type", tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"AdditionalData");
          break;
        case 'v': /* value */
          if(strcmp(p->name, "value") == 0)
	  {
  	     xmlNodeSetContent(cur, tmp_content);
             xmlFreeNode(p);
          }
          else
            badNode(p,"AdditionalData");
          break;
        default:
          badNode(p,"AdditionalData");
     }
     if(tmp_content != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(cur);

}

/*==================================================================
 * Name: newEnvironment
 *
 * Possible parameters 
 * 
 * SUB-ELEMENTS
 * char *env: one or more
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newEnvironment(char *env, ...)
{
  xmlNodePtr cur;
  va_list ap;
  char *p;
  char *e_value;

  if(env == NULL)
    return NULL;

  cur = xmlNewNode(NULL,"Environment");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newEnvironment: xmlNewNode(NULL,\"Environment\") failed\n");
    return NULL;
  }

  e_value = xmlEncodeEntitiesReentrant(doc,env);  /* escape special chars */
  xmlNewTextChild(cur,NULL,"env",e_value);
  xmlFree(e_value);

  va_start(ap, env);

  // step through the variable length argument list, and add text children
  for(p = env; p != NULL; p = va_arg(ap, char *))
  {
    // add the string to the argument list.  If something goes wrong,
    // return NULL.

    e_value = xmlEncodeEntitiesReentrant(doc,p);
    if(xmlNewTextChild(cur,NULL,"env",e_value) == NULL)
    {
       if(e_value != NULL)
         xmlFree(e_value);
        return NULL;
    }
    if(e_value != NULL)
       xmlFree(e_value);
  }
  return(cur);
}

/*==================================================================
 * Name: newArguments
 *
 * Possible parameters 
 * 
 * SUB-ELEMENTS
 * char *arg: one or more
 *
 * !Remember!: the last argument must be NULL.
 *===================================================================*/

xmlNodePtr newArguments(char *arg, ...)
{
  va_list ap;
  xmlNodePtr cur;
  char *p;
  char *e_value;

  if(arg == NULL)
    return NULL;

  cur = xmlNewNode(NULL,"Arguments");
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newArguments: xmlNewNode(NULL,\"Arguments\") failed\n");
    return NULL;
  }

  e_value = xmlEncodeEntitiesReentrant(doc,arg);
  xmlNewTextChild(cur,NULL,"arg",arg);
  xmlFree(e_value);

  va_start(ap, arg);

  // step through the variable length argument list, and add text children
  for(p = arg; p != NULL; p = va_arg(ap, char *))
  {
    // add the string to the argument list.  If something goes wrong,
    // return NULL.

    e_value = xmlEncodeEntitiesReentrant(doc,p);
    if(xmlNewTextChild(cur,NULL,"arg",e_value) == NULL)
    {
       if(e_value != NULL)
	  xmlFree(e_value);
       return NULL;
    }
    if(e_value != NULL)
       xmlFree(e_value);
  }
  return(cur);
}

/*==================================================================
 * Name: newSimpleElement
 *
 * Parameters 
 *  char *name : the parameter name
 *  char *value: the value of the parameter
 *===================================================================*/

xmlNodePtr newSimpleElement(char *name, char *value)
{
  xmlNodePtr cur;
  char *e_value;

  if((name == NULL) || (value == NULL))
    return NULL;

  cur = xmlNewNode(NULL,name);

  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newSimpleElement: xmlNewNode() failed\n");
    return NULL;
  }

  e_value = xmlEncodeEntitiesReentrant(doc,value); /* escape special chars */
  xmlNodeSetContent(cur,e_value);
  xmlFree(e_value);

  return(cur); 
}

/*==================================================================
 * Name: newAttribute
 *
 * Parameters 
 *  char *name : the parameter name
 *  char *value: the value of the parameter
 *
 * Note: This is the same function as newSimpleElement.  It was added to
 *       avoid confusion in building xml documents by following the
 *       nomenclature illustrated in the idmef-xml spec.  You may use
 *       newSimpleElement for adding new xml attributes if you so desire.
 *===================================================================*/

xmlNodePtr newAttribute(char *name, char *value)
{
  xmlNodePtr cur;
  char *e_value;
  
  if((name == NULL) || (value == NULL))
    return NULL;

  cur = xmlNewNode(NULL,name);
  if(cur == NULL)
  {
    fprintf(stderr,"idmefxml.c: newAttribute: xmlNewNode() failed\n");
    return NULL;
  }

  e_value = xmlEncodeEntitiesReentrant(doc,value);
  xmlNodeSetContent(cur,e_value);
  xmlFree(e_value);

  return(cur); 
}

/***********************************************************************
 *
 * These are functions that perform different operations on XML elements
 * that have been created with the above "new" functions.
 *
 ***********************************************************************/

/*===========================================================================
 * Name: addElement
 * Purpose: Add a child element to a parent.
 * Parameters:
 *           - parent : an xmlNodePtr to the parent
 *           - child  : an xmlNodePtr to the child
 *
 * Returns: A pointer to the child, or NULL if error
 *
 * Note: We're not sure what type of error checking should occur here.  There's
 *       not really a way to ensure that the child should exist under the
 *       parent without implementing part of the DTD.  My initial thought is
 *       to let the DTD validation catch any errors at run time.
 *==========================================================================*/

xmlNodePtr addElement(xmlNodePtr parent, xmlNodePtr child)
{
  return xmlAddChild(parent, child);
}

/*===========================================================================
 * Name: addElements
 * Purpose: Add a child elements to a parent.
 * Parameters:
 *           - parent : an xmlNodePtr to the parent
 *           - child  : an xmlNodePtr to a child
 *           - ...    : a variable size list of xmlNodePtrs to children
 *
 * Returns: the number of children added
 *
 * NOTE: The last parameter MUST be NULL.  Otherwise, the function will not
 *       know when to stop reading in the variable length parameter list, and
 *       bad things will happen.
 *==========================================================================*/

int addElements(xmlNodePtr parent, xmlNodePtr child, ...)
{
  xmlNodePtr p;
  va_list ap;
  int count = 0;

  if((parent == NULL) || (child == NULL)) //do we have at least two parameters?
    return 0;

  // cycle through the arguments
  va_start(ap, child);

  for(p = child; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     if(p->name != NULL)
     {
        xmlAddChild(parent, p);
        count++;
     }
  }
  va_end(ap);

  return(count);
}

/*===========================================================================
 * Name: setAttribute
 * Purpose: Set attributes to an element.
 * Parameters:
 *           - element : an xmlNodePtr to the element
 *           - attrib  : an xmlNodePtr to the attribute
 *
 * Returns: A pointer to the element, or NULL if error
 *==========================================================================*/

xmlNodePtr setAttribute(xmlNodePtr element, xmlNodePtr attrib)
{
  char *propName = NULL;
  xmlChar *tmp_content = NULL;
  propName = xmlStrdup(attrib->name);

  tmp_content = xmlNodeGetContent(attrib);

  if(xmlSetProp(element, propName, tmp_content))
  {
     xmlFreeNode(attrib);
     if(tmp_content != NULL)
        xmlFree(tmp_content);
     if(propName != NULL)
       xmlFree(propName);
     return element;
  }

  if(tmp_content != NULL)
     xmlFree(tmp_content);

  if(propName != NULL)
     xmlFree(propName);

  xmlFreeNode(attrib);
  return NULL;
}

/*===========================================================================
 * Name: setAttributes
 * Purpose: Set attributes to a element.
 * Parameters:
 *           - element : an xmlNodePtr to the element
 *           - attrib  : an xmlNodePtr to an attribute
 *           - ...     : a variable size list of xmlNodePtrs to attributes
 *
 * Returns: the number of attributes set
 *
 * NOTE: The last parameter MUST be NULL.  Otherwise, the function will not
 *       know when to stop reading in the variable length parameter list, and
 *       bad things will happen.
 *==========================================================================*/

int setAttributes(xmlNodePtr element, xmlNodePtr attrib, ...)
{
  xmlNodePtr p;
  va_list ap;
  char *propName = NULL;
  xmlChar *tmp_content = NULL;
  int count = 0;


  /* do we have at least two parameters? */
  if((element == NULL) || (attrib == NULL))
    return 0;

  // cycle through the arguments
  va_start(ap, attrib);

  for(p = attrib; p != NULL; p = va_arg(ap, xmlNodePtr))
  {
     tmp_content = xmlNodeGetContent(p);

     if(p->name != NULL)
     {
        propName = xmlStrdup(p->name);

        xmlSetProp(element, propName, tmp_content);
        count++;
     }
     xmlFreeNode(p);
 
     /* these values were copied in xmlSetProp, so free 'em */
     if(tmp_content != NULL)
        xmlFree(tmp_content);
     if(propName != NULL)
        xmlFree(tmp_content);
  }
  va_end(ap);

  return(count);
}

/*==========================================================================*
 * Name: getElement()
 * Purpose: Retrieve a pointer to the first child element directly under a
 *          parent element whose name matches the provided string.
 * Parameters:
 *           - parent : an xmlNodePtr to the parent
 *           - name   : a string representing the child's name
 *
 * Returns: A xmlNodePtr containing the element.
 *==========================================================================*/

xmlNodePtr getElement(xmlNodePtr parent, char *name)
{
  xmlNodePtr p;

  /* cycle through the children */
  for(p = parent->children; p != NULL; p = p->next)
  {
     if(strcmp(p->name, name) == 0)
        return p;
  }  
  return NULL;
}

 /***********************************************************************
  *
  * These are functions that perform different operations on XML documents.
  *
  ***********************************************************************/

/*==========================================================================*
 * Name: globalsInit()
 *
 * Purpose: Initialize global variables, and those structures we can reuse
 *          for each IDMEF Message.
 *
 * Parameters:
 *           - xml_version : The current version of XML
 *
 * Returns: 1 for success, and 0 otherwise
 *
 * NOTE: You MUST call this function before you start to create IDMEF messages
 *==========================================================================*/
int globalsInit(char *dtd_path)
{
  doc = NULL;

  xmlInitMemory();

  /* Initialize the validity context */

  vctxt = malloc(sizeof(xmlValidCtxt));

  if(vctxt == NULL)
  {
     fprintf(stderr,"idmefxml.c : globalsInit() : malloc fail!\n");
     return 0;
  }
  vctxt->userData = (void *) stderr;
  vctxt->error = (xmlValidityErrorFunc) fprintf;
  vctxt->warning = (xmlValidityWarningFunc) fprintf;

  /* initialize the DTD */
  dtd = xmlParseDTD(NULL,dtd_path);

  if(dtd == NULL)
  {
    fprintf(stderr,"idmefxml.c: createDoc(): problem parsing the DTD\n");
    return 0;
  }
  dtd->name = xmlStrdup((xmlChar*)"IDMEF-Message");

  return 1;
}

/*==========================================================================*
 * Name: createCurrentDoc()
 *
 * Purpose: Creates the a new xmlDoc and sets it as the current (doc).
 *
 * Parameters:
 *           - xml_version : The current version of XML
 *
 * Returns: 1 for success, and 0 otherwise
 *==========================================================================*/
int createCurrentDoc(const char *xml_version)
{
  xmlDocPtr thisDoc;
  thisDoc = createDoc(xml_version);
  
  if(setCurrentDoc(thisDoc))
     return 1;
  return 0;
}

/*==========================================================================*
 * Name: createDoc()
 * Purpose: Creates a new xmlDoc, initializes the root element, and links
 *          to the DTD.
 *
 * Parameters:
 *           - xml_version : The current version of XML
 *
 * Returns: an xmlDocPtr to the newly created doc, or NULL for error.
 *==========================================================================*/

xmlDocPtr createDoc(const char *xml_version)
{
  xmlDocPtr newDoc;
  xmlNodePtr root, oldRoot;
 
  /* Initialize the document and root element */

  newDoc = xmlNewDoc(xml_version);
  
  if(newDoc == NULL)
  {
    fprintf(stderr,"idmefxml.c: createDoc(): problem creating new xml doc\n");
    return 0;
  }

  root = xmlNewNode(NULL,"IDMEF-Message");

  if(root == NULL)
  {
    fprintf(stderr,"idmefxml.c: newIDMEFDoc(): problem creating the root node\n");
    return NULL;
  }

  oldRoot = xmlDocSetRootElement(newDoc,root); 

  if(oldRoot != NULL)
    xmlFreeNode(oldRoot);

  /* Point the new document to the dtd */

  newDoc->intSubset = dtd;

  return newDoc;
}

/*==========================================================================*
 * Name: setCurrentDoc()
 * 
 * Purpose : Sets the current doc and free's the old doc if need be.
 *
 * Parameters:
 *         - newDoc : An xmlDocPtr to the doc to be set as the current
 *
 * Returns: 1 if success, 0 otherwise
 *==========================================================================*/
int setCurrentDoc(xmlDocPtr newDoc)
{
   if(newDoc == NULL)
   {
      fprintf(stderr,"idmefxml.c : setCurrentDoc() - NULL parameter newDoc\n");
      return 0;
   }
   
   if(doc != NULL)      /* was there a current doc already? */
   { 
      clearCurrentDoc();
   }

   doc = newDoc;
   dtd->doc = doc;

   return 1;
}

/*==========================================================================*
 * Name: getCurrentDoc()
 * Purpose: gets the current xmlDoc (the global "doc").
 *
 * Returns: an xmlDocPtr to the current doc.
 *==========================================================================*/
xmlDocPtr getCurrentDoc()
{
  return doc;
}

/*==========================================================================*
 * Name: clearCurrentDoc()
 *
 * Purpose: Free the current xmlDoc, including all it's structures, except
 *          for the DTD.  The DTD is unlinked first, as it is re-usable.
 *
 *==========================================================================*/
void clearCurrentDoc()
{
    if(doc == NULL)
      return;

    dtd->doc = NULL;
    doc->intSubset = NULL;
    xmlFreeDoc(doc);    
    doc = NULL;    
    xmlCleanupParser();
}

/*==========================================================================*
 * Name: copyCurrentDoc()
 *
 * Purpose: Create a copy of the current xmlDoc.  This copies everything
 *          _except_ the DTD, as the DTD is re-usable, and we don't need
 *          extraneous copies running amok.
 *
 * Returns:
 *        - docCopy : A copy of the current xml doc
 *
 * NOTE: The xmlDoc copy is NOT linked to the DTD.  If you wish it to be,
 *       setCurrentDoc must be called on the document.
 *==========================================================================*/
xmlDocPtr copyCurrentDoc()
{
  xmlDocPtr docCopy;

  doc->intSubset = NULL;        /* unlink the dtd */
  docCopy = xmlCopyDoc(doc,1);  /* recursively copy that bad boy */
  doc->intSubset = dtd;         /* relink the dtd */

  return docCopy;
}

/*==========================================================================*
 * Name: resetCurrentDoc()
 *
 * Purpose: Reset the current document by clearing it and re-initializing
 *          it.
 *
 * Returns: 1 for success, and 0 otherwise
 *==========================================================================*/
int resetCurrentDoc()
{
  xmlChar *theVersion;

  theVersion = xmlStrdup(doc->version);

  clearCurrentDoc();
 
  if(createCurrentDoc(theVersion))
  {
     if(theVersion != NULL)
        xmlFree(theVersion);
     return 1;
  }

  if(theVersion != NULL)
     xmlFree(theVersion);

  return 0;  
}

/*==========================================================================*
 * Name: validateCurrentDoc()
 *
 * Purpose: Validates the current document
 *
 * Returns: 1 for success, and 0 otherwise
 *==========================================================================*/
int validateCurrentDoc()
{
  if(!xmlValidateDtd(vctxt,doc,doc->intSubset))  
    return 0;
  return 1;
}

/*==========================================================================*
 * Name: printCurrentMessage()
 *
 * Purpose: Prints the current message to the FILE * provided
 *==========================================================================*/
void printCurrentMessage(FILE *f)
{
  xmlElemDump(f,doc,doc->children);
}

/***********************************************************************
 *
 *               Miscellaneous time format functions.
 *
 ***********************************************************************/

/*==========================================================================*
 * Name: currentNtpTimestamp()
 *
 * Purpose: Creates a NTP timestamp with the current time
 *
 * Returns: The current NTP timestamp as a char *
 *==========================================================================*/
char *currentNtpTimestamp()
{
   char *ntp_timestamp;
   time_t curr_time;

   curr_time = time(NULL);

   ntp_timestamp = calloc(MAX_NTP_TIMESTAMP_SIZE,sizeof(char));
   sprintf(ntp_timestamp,"0x%lx.0x0",(long)curr_time);

   return ntp_timestamp;
}

/*==========================================================================*
 * Name: currentIDMEF_Time()
 *
 * Purpose: Formats the current time as specified by the IDMEF XML
 *          specification version 1. HH:MM:SS
 *
 * Returns: The formated time as a char *
 *==========================================================================*/
char *currentIDMEF_Time()
{
   char *time_str;
   time_t curr_time;
   struct tm *loc_time;

   curr_time = time(NULL);
   loc_time = localtime(&curr_time);

   time_str = calloc(MAX_TIME_SIZE,sizeof(char));
   strftime(time_str,MAX_TIME_SIZE,"%T",loc_time);
 
   return time_str;
}

/*==========================================================================*
 * Name: currentIDMEF_Date()
 *
 * Purpose: Formats the current date as specified by the IDMEF XML
 *          specification version 1. YYYY-MM-DD
 *
 * Returns: The formated date as a char *
 *==========================================================================*/
char *currentIDMEF_Date()
{
   char *date_str;
   time_t curr_time;
   struct tm *loc_time;

   curr_time = time(NULL);
   loc_time = localtime(&curr_time);

   date_str = calloc(MAX_DATE_SIZE,sizeof(char));
   strftime(date_str,MAX_DATE_SIZE,"%Y-%m-%d",loc_time);
 
   return date_str;
}

/*==========================================================================*
 * Name: convertNtpTimestamp()
 *
 * Purpose: Creates a NTP timestamp with the provided time
 *
 * Returns: The provided NTP timestamp as a char *
 *==========================================================================*/
char *convertNtpTimestamp(const time_t *curr_time)
{
   char *ntp_timestamp;

   ntp_timestamp = calloc(MAX_NTP_TIMESTAMP_SIZE,sizeof(char));
   sprintf(ntp_timestamp,"0x%lx.0x0",(long)*curr_time);

   return ntp_timestamp;
}

/*==========================================================================*
 * Name: convertTime()
 *
 * Purpose: Formats the provided time as specified by the IDMEF XML
 *          specification version 1. HH:MM:SS
 *
 * Returns: The formated time as a char *
 *==========================================================================*/
char *convertTime(const time_t *curr_time)
{
   char *time_str;
   struct tm *loc_time;

   loc_time = localtime(curr_time);

   time_str = calloc(MAX_TIME_SIZE,sizeof(char));
   strftime(time_str,MAX_TIME_SIZE,"%T",loc_time);
 
   return time_str;
}

/*==========================================================================*
 * Name: convertDate()
 *
 * Purpose: Formats the provided date as specified by the IDMEF XML
 *          specification version 1. YYYY-MM-DD
 *
 * Returns: The formated date as a char *
 *==========================================================================*/
char *convertDate(const time_t *curr_time)
{
   char *date_str;
   struct tm *loc_time;

   loc_time = localtime(curr_time);

   date_str = calloc(MAX_DATE_SIZE,sizeof(char));
   strftime(date_str,MAX_DATE_SIZE,"%Y-%m-%d",loc_time);
 
   return date_str;
}

/***********************************************************************
 *
 *                   Other utility functions
 *
 ***********************************************************************/

/*==========================================================================*
 * Name: getStoredAlertID()
 *
 * Purpose: Opens the specified file, and attempts to extract the stored
 *          alert id.
 *
 * Returns: the alert id as an unsigned long integer, 1 if the file didn't
 *          contain an alert id, or 0 for error.
 *==========================================================================*/
ulong getStoredAlertID(char *f_name)
{
  FILE *f;
  char *idmef_alertid_string;
  ulong aid;

  /* Open the alert ID file and extract the stored alert ID if it exists*/
  f = fopen(f_name,"a+");
  if (!f) 
  {
     fprintf(stderr,"getStoredAlertID: IDMEF alert ID file open error (%s)\n",f_name);
     return 0;
  }

  /** Now seek back to the beginning... Nice little hack **/
  rewind(f);
    
  idmef_alertid_string = (char *) calloc(MAX_ALERTID_BUFFER_SIZE,sizeof(char));

  if(fgets(idmef_alertid_string, MAX_ALERTID_BUFFER_SIZE + 1, f) == NULL)
  {
    //     fprintf(stderr,"getStoredAlertID: Stored alert ID not found in %s, continuing with alert ID = 1\n", f_name);
     aid = 1;
  }
  else
     aid = (ulong) atol(idmef_alertid_string);
  
  return aid;

  fclose(f); /* done for now */
}

/*==========================================================================*
 * Name: saveAlertID()
 *
 * Purpose: Opens the specified file, and attempts to extract the stored
 *          alert id.
 *
 * Returns: 1 for success or 0 for error.
 *==========================================================================*/
int saveAlertID(ulong aid, char *f_name)
{
   char *temp;
   FILE *f;
 
   temp = (char *) calloc(MAX_ALERTID_BUFFER_SIZE,sizeof(char));
   temp = ulongToString(aid);

   f = fopen(f_name,"w");
   if (!f) 
   {
      fprintf(stderr,"saveAlertID: IDMEF alert ID file open error (%s)\n",f_name);
      return 0;
   }

   if(fputs(ulongToString(aid), f) < 0)
   {
     fprintf(stderr,"saveAlertID: Error writting alert id (%lu) to %s\n", aid, f_name);
     return 0;
   }
   fputc('\n',f);
   fflush(f);
   fclose(f);

   return 1;
}

/*==========================================================================*
 * Name: badNode()
 * Purpose: Prints a message to stderr with the invalid node name, and the
 *          parent it was attempted to be linked to.
 *
 * Parameters:
 *           - theNode : the invalid node
 *           - parent  : the would-be parent node's name
 *==========================================================================*/
void badNode(xmlNodePtr theNode, char *parent)
{
   fprintf(stderr, "idmefxml.c: Invalid xmlNode \"%s\" for %s\n",
          theNode->name, parent);
   xmlFreeNode(theNode);
}

/*==========================================================================*
 * Name: intToString()
 * Purpose: Converts a signed integer value to it's ASCII representation.
 *
 * Parameters:
 *           - int_val : the value to be converted
 *
 * Returns: The ASCII representation
 *==========================================================================*/
char * intToString(int int_val)
{
  int ret, tmp;
  char *string_val;

  if(int_val > 0)
    tmp = floor(log10(int_val)) + 2;
  else
     if(int_val < 0)
        tmp = floor(log10(-1 * int_val)) + 3;
     else
        tmp = 2;

  string_val = xmlMalloc(tmp);
  //  string_val = calloc(tmp,sizeof(char));
  if(string_val == NULL)
     return NULL;
  ret = sprintf(string_val,"%d",int_val);

  if(ret > -1 && ret <= tmp)
     return string_val;
  
  return NULL;
}

/*==========================================================================*
 * Name: ulongToString()
 * Purpose: Converts an unsigned value to it's ASCII representation.
 * 
 * Parameters:
 *           - ulong_val : the value to be converted
 *
 * Returns: The ASCII representation
 *
 * Note: This function was custom made for converting ulong alertID's to
 *       strings.
 *==========================================================================*/
char * ulongToString(ulong ulong_val)
{
  int ret;
  ulong tmp;
  char *string_val;

  if(ulong_val == 0)
    tmp = 2;
  else
    tmp = floor(log10(ulong_val)) + 2;
 
  string_val = calloc(tmp,sizeof(char));
  if(string_val == NULL)
     return NULL;
  ret = sprintf(string_val,"%lu",ulong_val);

  if(ret > -1 && ret <= tmp)
     return string_val;

  return NULL;
}
