/************************************************************************
 * libidmef: A library to create IDMEF messages in XML format.
 * Author: Joe McAlerney, Silicon Defense, (joey@SiliconDefense.com)
 *
 * Copyright (c) 2000 by Silicon Defense (http://www.silicondefense.com/)
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 **************************************************************************/

/***************************************************************
* This is an implementation of the IDMEF XML DTD described in
* draft-ietf-idwg-idmef-xml-01.txt
****************************************************************/

#ifndef IDMEF_XML_H
#define IDMEF_XML_H

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/valid.h>
#include <libxml/entities.h>
#include <libxml/xmlmemory.h>
#include <libxml/debugXML.h>

#include "time.h"

#ifdef WIN32
typedef unsigned long ulong; /* windows uses ULONG, but not ulong */
#endif

#define IDMEF_MESSAGE_VERSION "0.1"
#define MAX_ALERTID_BUFFER_SIZE 11   /* ulong (2^32 + 1) */
#define MAX_DATE_SIZE           11
#define MAX_TIME_SIZE           12
#define MAX_NTP_TIMESTAMP_SIZE  15

/*******************************************
 * Global variables
 *******************************************/

xmlDocPtr doc;
xmlDtdPtr dtd;
xmlValidCtxtPtr vctxt;

/*=================================================
 *  Enumeration definitions
 *=================================================*/

/* I still don't like the naming convention for these enum's.  Unfortunately,
   the same enum constant can not be used more than once, so we can not
   model the IDMEF XML DTD perfectly.  Maybe enum's are not the answer. */

typedef enum _impact { 
                        IMPACT_UNKNOWN   = 0,
                        NOT_SUSPICIOUS   = 1,
                        BAD_UNKNOWN      = 2,
                        ATTEMPTED_RECON  = 3,
                        SUCCESSFUL_RECON = 4,
			ATTEMPTED_DOS    = 5,
			SUCCESSFUL_DOS   = 6,
			ATTEMPTED_USER   = 7,
			SUCCESSFUL_USER  = 8,
			ATTEMPTED_ADMIN  = 9,
			SUCCESSFUL_ADMIN = 10
} impact;

typedef enum _method {
                        METHOD_UKNNOWN   = 0,
			KNOWLEDGE        = 1,
			BEHAVIOR         = 2,
			CORRELATION      = 3,
			METHOD_POLICY    = 4
} method;

typedef enum _spoofed {
                        SPOOFED_UNKNOWN  = 0,
			SPOOFED_YES      = -1,
			SPOOFED_NO       = 1
} spoofed;

typedef enum _decoy {
                        DECOY_UNKNOWN    = 0,
		        DECOY_YES        = -1,
		        DECOY_NO         = 1
} decoy;

typedef enum _address_category {
                        ADDRESS_CATEGORY_UNKNOWN = 0,
			IPV4_ADDR_HEX    = 1,
			IPV4_ADDR        = 2,
			IPV4_NET         = 3,
			IPV6_ADDR        = 4,
			IPV6_NET         = 5,
			MAC              = 6,
			ATM              = 7,
			SNA              = 8,
			E_MAIL           = 20,
			LOTUS_NOTES      = 21,
			VM               = 22
} address_category;

typedef enum _origin {
                        UNKNOWN_ORIGIN   = 1,
			BUGTRAQID        = 2,
			CVE              = 3,
			ORIGIN_POLICY    = 4
} origin;

typedef enum _node_category {
                        NODE_CATEGORY_UNKNOWN = 0,
			DNS              = 1,
			NIS              = 2,
			NISPLUS          = 3,
			WFW              = 4,
			NT               = 5,
			ADS              = 6,
			NDS              = 7,
			AFS              = 8,
			KERBEROS         = 9,
			DFS              = 10,
			CODA             = 11
} node_category;

typedef enum _user_category {
                        USER_CATEGORY_UNKNOWN = 0,
			OS_DEVICE        = 1,
			APPLICATION      = 2
} user_category;

typedef enum _type {
                        TYPE_UNKNOWN     = 0,
			BOOLEAN          = 1,
			BYTE             = 2,
			CHARACTER        = 3,
			DATE             = 4,
			INTEGER          = 5,
			REAL             = 6,
			STRING           = 7,
			TIME             = 8
} type;

/*=================================================*
 *           Data structure definitions            *
 *=================================================*/

/* Define the data structures that will represent classes as defined in
   the Intrusion Detection Exchange Format Data Model

   I really don't like having to append IDMEF_ in front of each of these
   structures, but to avoid name collisions I am going to do it.       

   NOTE: These data structures are not currently being used by any functions
         in libidmef.  If it is found that there is a need or want for them,
         they will be used.
*/

/*==================
   Simple Elements  
 ===================*/

/*
 * 6.6.2 The Arguments Element
 */
typedef struct _Arguments {
  char **arg;
} IDMEF_Arguments;

/*
 * 6.6.9 The Environment Element
 */
typedef struct _Environment {
  char **env;
} IDMEF_Environment;

/*==================================
   Providing Additional Information  
 ===================================*/

/*
 * 6.7.1 The AdditionalData Element
 */
typedef struct _AdditionalData {
  char *meaning;
  char *type;
  char *value;                   // the data itself
} IDMEF_AdditionalData;

/*===========================================
   Low-Level Entity Identification Elements  
 ============================================*/

/*
 * 6.5.1 The Address Element
 */
typedef struct _Address {
  char *ident;
  char *category;
  char *address;
  char *netmask;
} IDMEF_Address;

/*
 * 6.5.2 The Classification Element
 */
typedef struct _Classification {
  char *origin;
  char *name;
  char *url;
} IDMEF_Classification;

/*
 * 6.5.3 The Node Element
 */
typedef struct _Node {
  char *ident;
  char *category;
  char *name;
  char *location;
  IDMEF_Address *node_Address;
} IDMEF_Node;

/*
 * 6.5.4 The Process Element
 */
typedef struct _Process {
  char *ident;
  char *name;
  int pid;
  char *path;
  IDMEF_Arguments *process_Arguments;
  IDMEF_Environment *process_Environment;
} IDMEF_Process;

/*
 * 6.5.5.1 The SNMPService Element
 */
typedef struct _SNMPService {
  char *ident;
  char *oid;
  char *community;
  char *command;
} IDMEF_SNMPService;

/*
 * 6.5.5.2 The WebService Element
 */
typedef struct _WebService {
  char *ident;
  char *url;
  char *cgi;
  char *method;
  IDMEF_Arguments *webservice_Arguments;
} IDMEF_WebService;

/*
 * 6.5.5 The Service Element
 */
typedef struct _Service {
  char *ident;
  char *name;
  int dport;
  int sport;
  char *protocol;
  char *portlist;
  IDMEF_SNMPService *service_SNMPService;
  IDMEF_WebService *service_WebService;
} IDMEF_Service;

/*
 * 6.5.6 The User Element
 */
typedef struct _User {
  char *ident;
  char *category;
  char *name;
  int uid;
  char *group;
  int gid;
  char *serial;
  IDMEF_Address *user_Address;
} IDMEF_User;

/*============================================
   High-Level Entity Identification Elements  
 =============================================*/

/*
 * 6.4.1 The Analyzer Element
 */
typedef struct _Analyzer {
  char *ident;
  IDMEF_Node *analyzer_Node;
  IDMEF_Process *analyzer_Process;
} IDMEF_Analyzer;

/*
 * 6.4.2 The Source Element
 */
typedef struct _Source {
  char *sourceid;
  char *spoofed;
  IDMEF_Node *source_Node;
  IDMEF_User *source_User;
  IDMEF_Process *source_Process;
} IDMEF_Source;

/*
 * 6.4.3 The Target Element
 */
typedef struct _Target {
  char *targetid;
  char *decoy;
  IDMEF_Node *target_Node;
  IDMEF_User *target_User;
  IDMEF_Process *target_Process;
  IDMEF_Service *target_Service;
} IDMEF_Target;

/*========================
       Time Elements  
 =========================*/

/*
 * 6.3.2 The DetectTime Element
 */
typedef struct _DetectTime {
  int offset;
  char *date;
  char *time;
} IDMEF_DetectTime;

/*
 * 6.3.3 The AnalyzerTime Element
 */
typedef struct _AnalyzerTime {
  int offset;
  char *date;
  char *time;
} IDMEF_AnalyzerTime;

/*
 * 6.3.1 The Time Element
 */
typedef struct _Time {
  int offset;
  char *ntpstamp;
  char *date;
  char *time;
  IDMEF_DetectTime *time_DetectTime;
  IDMEF_AnalyzerTime *time_AnalyzerTime;
} IDMEF_Time;

/*========================
   Message Type Elements  
 =========================*/

/*
 * 6.2.1.1 The CorrelationAlert Element
 */
typedef struct _CorrelationAlert {
  char **alertid;
} IDMEF_CorrelationAlert;

/*
 * 6.2.1.2 The OverflowAlert Element
 */
typedef struct _OverflowAlert {
  /* NOTE - The draft doesn't call for an alertID here,
     but there probably should be                */
  char *program;
  int size;
  char *buffer;          /* SHOULD BE AN ARRAY OF BYTES... */
} IDMEF_OverflowAlert;

/*
 * 6.2.1.3 The ToolAlert Element
 */
typedef struct _ToolAlert {
  char *name;
  char *command;
  char **alertid;
} IDMEF_ToolAlert;

/*
 * 6.2.1 The Alert Element
 */

typedef struct _Alert {
  char *alertid;
  char *impact;
  char *version;
  IDMEF_Time *alert_Time;
  IDMEF_Analyzer *alert_Analyzer;
  IDMEF_Classification *alert_Classification;
  IDMEF_Source *alert_Source;
  IDMEF_Target *alert_Target;

  IDMEF_ToolAlert *tAlert;
  IDMEF_OverflowAlert *ofAlert;
  IDMEF_CorrelationAlert *cAlert;
  IDMEF_AdditionalData *alert_AdditionalData;
} IDMEF_Alert;


/*
 * 6.2.2 The Heartbeat Element
 */
typedef struct _Heartbeat {
  char *heartbeatid;
  IDMEF_Time *heartbeat_Time;
  IDMEF_Analyzer *heartbeat_Analyzer;
  IDMEF_AdditionalData *heartbeat_AdditionalData;
} IDMEF_Heartbeat;

/*=============================
   IDMEF-Message Root Element
 ==============================*/
/*
 * 6.1 The IDMEF-Message Root Element
 */
typedef struct _IDMEF_Message {
  char *version;
  IDMEF_Alert *IDMEF_Message_Alert;
  IDMEF_Heartbeat *IDMEF_Message_Heartbeat;
} IDMEF_Message;

/*=================================================*
 *               Function Prototypes               *
 *=================================================*/

/*=============================================================================
   New Element functions  

   Sub-elements/attributes that are not required, or additonal instances
   of the required ones, will be appended at the end of the parameter string -
   denoted by "..."
 ============================================================================*/

xmlNodePtr newIDMEF_Message(xmlNodePtr, ...);
xmlNodePtr newAlert(xmlNodePtr, ...);
xmlNodePtr newHeartbeat(xmlNodePtr, ...);
xmlNodePtr newToolAlert(xmlNodePtr, ...);
xmlNodePtr newOverflowAlert(xmlNodePtr, ...);
xmlNodePtr newCorrelationAlert(xmlNodePtr, ...);
xmlNodePtr newTime(xmlNodePtr, ...);
xmlNodePtr newAnalyzerTime(xmlNodePtr, ...);
xmlNodePtr newDetectTime(xmlNodePtr, ...);
xmlNodePtr newTarget(xmlNodePtr, ...);
xmlNodePtr newSource(xmlNodePtr, ...);
xmlNodePtr newAnalyzer(xmlNodePtr, ...);
xmlNodePtr newUser(xmlNodePtr, ...);
xmlNodePtr newService(xmlNodePtr, ...);
xmlNodePtr newWebService(xmlNodePtr, ...);
xmlNodePtr newSNMPService(xmlNodePtr, ...);
xmlNodePtr newProcess(xmlNodePtr, ...);
xmlNodePtr newNode(xmlNodePtr, ...);
xmlNodePtr newClassification(xmlNodePtr, ...);
xmlNodePtr newAddress(xmlNodePtr, ...);
xmlNodePtr newAdditionalData(xmlNodePtr, ...);
xmlNodePtr newEnvironment(char *, ...);
xmlNodePtr newArguments(char *, ...);
xmlNodePtr newSimpleElement(char *, char *);
xmlNodePtr newAttribute(char *, char *);

/*-----------------------------------------------------------------------*
 * These are functions that perform different operations on XML elements
 * that have been created with the above "new" functions.
 *-----------------------------------------------------------------------*/

xmlNodePtr addElement(xmlNodePtr, xmlNodePtr);
int addElements(xmlNodePtr, xmlNodePtr, ...);
xmlNodePtr setAttribute(xmlNodePtr, xmlNodePtr);
int setAttributes(xmlNodePtr, xmlNodePtr, ...);
xmlNodePtr getElement(xmlNodePtr, char *);

/*-------------------------------------------------------------------------*
 * These are functions that perform different operations on XML documents.
 *-------------------------------------------------------------------------*/

int globalsInit(char *);
int createCurrentDoc(const char *);
xmlDocPtr createDoc(const char *);
int setCurrentDoc(xmlDocPtr);
xmlDocPtr getCurrentDoc();
void clearCurrentDoc();
xmlDocPtr copyCurrentDoc();
int resetCurrentDoc();
int validateCurrentDoc();
void printCurrentMessage(FILE *);

/*---------------------------------------
 * Miscellaneous time format functions.
 *---------------------------------------*/

char *currentNtpTimestamp();
char *currentIDMEF_Time();
char *currentIDMEF_Date();
char *convertNtpTimestamp(const time_t *);
char *convertTime(const time_t *);
char *convertDate(const time_t *);

/*--------------------------------
 * Other utility functions
 *-------------------------------*/

unsigned long getStoredAlertID(char *);
int saveAlertID(unsigned long, char *);
void badNode(xmlNodePtr, char *);
char * intToString(int);
char * ulongToString(unsigned long);

/*=============================================================================
   Functions that convert xml node elements to c structures  

   These have not been implemented yet, as it is not known whether
   they would be useful.
 ============================================================================*/
/*
IDMEF_Message *  convertIDMEF_Message(xmlNodePtr);
IDMEF_Alert * convertAlert(xmlNodePtr);
IDMEF_Heartbeat * convertHeartbeat(xmlNodePtr);
IDMEF_ToolAlert * convertToolAlert(xmlNodePtr);
IDMEF_OverflowAlert * convertOverflowAlert(xmlNodePtr);
IDMEF_CorrelationAlert * convertCorrelationAlert(xmlNodePtr);
IDMEF_Time * convertTime(xmlNodePtr);
IDMEF_AnalyzerTime * convertAnalyzerTime(xmlNodePtr);
IDMEF_DetectTime * convertDetectTime(xmlNodePtr);
IDMEF_Target * convertTarget(xmlNodePtr);
IDMEF_Source * convertSource(xmlNodePtr);
IDMEF_Analyzer * convertAnalyzer(xmlNodePtr);
IDMEF_User * convertUser(xmlNodePtr);
IDMEF_Service * convertService(xmlNodePtr);
IDMEF_WebService * convertWebService(xmlNodePtr);
IDMEF_SNMPService * convertSNMPService(xmlNodePtr);
IDMEF_Process * convertProcess(xmlNodePtr);
IDMEF_Node * convertNode(xmlNodePtr);
IDMEF_Classification * convertClassification(xmlNodePtr);
IDMEF_Address * convertAddress(xmlNodePtr);
IDMEF_AdditionalData * convertAdditionalData(xmlNodePtr);
IDMEF_Environment * convertEnvironment(xmlNodePtr);
IDMEF_Arguments * convertArguments(xmlNodePtr);
*/

#endif IDMEF_XML_H
